/*******************************************************************************
SiJaPP - Simple Java PreProcessor
Copyright (C) 2003  Manuel Linsmayer

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*******************************************************************************/


package sijapp;


import java.util.Vector;


public class Scanner {


  // Alphabetic character test
  private static boolean isAlpha(char c) {
    return (((c >= 'A') && (c <= 'Z')) || ((c >= 'a') && (c <= 'z')));
  }


  // Numeric character test
  private static boolean isNumeric(char c) {
    return ((c >= '0') && (c <= '9'));
  }


  // Dot test
  private static boolean isDot(char c) {
    return (c == '.');
  }


  // Whitespace test
  private static boolean isWhitespace(char c) {
    return ((c == ' ') || (c == '\t'));
  }


  // Backslash test
  private static boolean isBackslash(char c) {
    return (c == '\\');
  }


  // Quote test
  private static boolean isQuote(char c) {
    return (c == '"');
  }


  // Pound test
  private static boolean isPound(char c) {
    return (c == '#');
  }


  // Underscore
  private static boolean isUnderscore(char c) {
    return (c == '_');
  }


  // Scanner
  public static Token[] scan(String s) throws SijappException {

    // Look for SiJaPP statement
    if (s.indexOf("#sijapp") == -1) {
      return (new Token[0]);
    }

    // Crop anything to the left
    StringBuffer buf = new StringBuffer(s.substring(s.indexOf("#sijapp")) + " ");

    // Vector
    Vector tokens = new Vector();

    // Required string buffers
    StringBuffer ident = null;
    StringBuffer string = null;

    // State variable
    int state = 0;

    // Exit flag
    boolean exit = false;

    // Scanner main loop
    do {

      // Get next character
      char ch = buf.charAt(0);
      buf.deleteCharAt(0);

      // Deterministic finite automaton, see README file
      switch (state) {
        case 0:
          if (Scanner.isPound(ch)) {
            ident = new StringBuffer();
            state = 1;
            break;
          }
          else if (Scanner.isAlpha(ch) || Scanner.isUnderscore(ch)) {
            ident = new StringBuffer();
            ident.append(ch);
            state = 3;
            break;
          }
          else if (Scanner.isDot(ch)) {
            state = 4;
            break;
          }
          else if (Scanner.isQuote(ch)) {
            string = new StringBuffer();
            state = 5;
            break;
          }
          else if (Scanner.isWhitespace(ch)) {
            state = 8;
            break;
          }
          else {
            throw (new SijappException("Syntax error"));
          }
        case 1:
          if (Scanner.isAlpha(ch) || Scanner.isUnderscore(ch)) {
            ident.append(ch);
            state = 2;
            break;
          }
          else {
            tokens.add(new Token(Token.T_MAGIC_END, null));
            exit = true;
            buf.insert(0, ch);
            state = 0;
            break;
          }
        case 2:
          if (Scanner.isAlpha(ch) || Scanner.isNumeric(ch) || Scanner.isUnderscore(ch)) {
            ident.append(ch);
            state = 2;
            break;
          }
          else {
            if (ident.toString().equals("sijapp")) {
              tokens.add(new Token(Token.T_MAGIC_BEGIN, null));
              buf.insert(0, ch);
              state = 0;
              break;
            }
            else {
              throw (new SijappException("Syntax error"));
            }
          }
        case 3:
          if (Scanner.isAlpha(ch) || Scanner.isNumeric(ch) || Scanner.isUnderscore(ch)) {
            ident.append(ch);
            state = 3;
            break;
          }
          else {
            if (ident.toString().equals("echo")) {
              tokens.add(new Token(Token.T_CMD_ECHO, null));
            }
            else if (ident.toString().equals("exit")) {
              tokens.add(new Token(Token.T_CMD_EXIT, null));
            }
            else if (ident.toString().equals("target")) {
              tokens.add(new Token(Token.T_CMD_TARGET, null));
            }
            else if (ident.toString().equals("if")) {
              tokens.add(new Token(Token.T_CMD_TARGET_IF, null));
            }
            else if (ident.toString().equals("elseif")) {
              tokens.add(new Token(Token.T_CMD_TARGET_ELSEIF, null));
            }
            else if (ident.toString().equals("unless")) {
              tokens.add(new Token(Token.T_CMD_TARGET_UNLESS, null));
            }
            else if (ident.toString().equals("else")) {
              tokens.add(new Token(Token.T_CMD_TARGET_ELSE, null));
            }
            else if (ident.toString().equals("end")) {
              tokens.add(new Token(Token.T_CMD_TARGET_END, null));
            }
            else {
              tokens.add(new Token(Token.T_IDENT, ident.toString()));
            }
            buf.insert(0, ch);
            state = 0;
            break;
          }
        case 4:
          tokens.add(new Token(Token.T_CMDSEP, null));
          buf.insert(0, ch);
          state = 0;
          break;
        case 5:
          if (Scanner.isQuote(ch)) {
            state = 6;
            break;
          }
          else if (Scanner.isBackslash(ch)) {
            state = 7;
            break;
          }
          else {
            string.append(ch);
            state = 5;
            break;
          }
        case 6:
          tokens.add(new Token(Token.T_STRING, string.toString()));
          buf.insert(0, ch);
          state = 0;
          break;
        case 7:
          if (Scanner.isQuote(ch) || Scanner.isBackslash(ch)) {
            string.append(ch);
            state = 5;
            break;
          }
          else {
            throw (new SijappException("Syntax error"));
          }
        case 8:
          buf.insert(0, ch);
          state = 0;
          break;
      }

    } while (!exit && (buf.length() > 0));

    // Return tokens as array
    Token[] ret = new Token[tokens.size()];
    tokens.copyInto(ret);
    return (ret);

  }


  /****************************************************************************/
  /****************************************************************************/
  /****************************************************************************/


  public static class Token {


    // Tokens
    public static final int T_MAGIC_BEGIN       = 1;
    public static final int T_MAGIC_END         = 2;
    public static final int T_CMDSEP            = 3;
    public static final int T_IDENT             = 4;
    public static final int T_CMD_ECHO          = 5;
    public static final int T_CMD_EXIT          = 6;
    public static final int T_CMD_TARGET        = 7;
    public static final int T_CMD_TARGET_IF     = 8;
    public static final int T_CMD_TARGET_UNLESS = 9;
    public static final int T_CMD_TARGET_ELSEIF = 10;
    public static final int T_CMD_TARGET_ELSE   = 11;
    public static final int T_CMD_TARGET_END    = 12;
    public static final int T_STRING            = 13;


    /**************************************************************************/


    // Token type
    private int type;


    // Token value
    private Object value;


    // Constructor
    public Token(int type, Object value) {
      this.type = type;
      this.value = value;
    }


    // Returns the token type
    public int getType() {
      return (this.type);
    }


    // Returns the token value
    public Object getValue() {
      return (this.value);
    }


  }


}
